/*******************************************************************************
 * Copyright (c) 2006, 2014 Tom Schindl and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *     Tom Schindl <tom.schindl@bestsolution.at> - initial API and implementation
 *     Florian Potschka <signalrauschen@gmail.com> - Bug 260061
 *     Alexander Ljungberg <siker@norwinter.com> - Bug 260061
 *     Jeanderson Candido <http://jeandersonbc.github.io> - Bug 414565
 *     Alexander Mark Diewald <diewald@fortiss.org>
 *******************************************************************************/
package org.fortiss.tooling.base.ui.annotation.labelprovider;

import static org.eclipse.jface.resource.JFaceResources.getImageRegistry;

import org.eclipse.jface.resource.JFaceResources;
import org.eclipse.jface.viewers.ColumnLabelProvider;
import org.eclipse.jface.viewers.ColumnViewer;
import org.eclipse.jface.viewers.TableViewer;
import org.eclipse.jface.viewers.ViewerCell;
import org.eclipse.jface.viewers.ViewerRow;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.MouseEvent;
import org.eclipse.swt.events.MouseListener;
import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.graphics.GC;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.graphics.ImageData;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.graphics.Rectangle;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.swt.widgets.Table;
import org.eclipse.swt.widgets.TableItem;
import org.fortiss.tooling.base.annotation.AnnotationEntry;
import org.fortiss.tooling.base.model.element.IAnnotatedSpecification;
import org.fortiss.tooling.base.ui.annotation.view.IAnnotationViewPart;

/**
 * <p>
 * A {@link ColumnLabelProvider} for annotations that displays checkboxes indicating the value of
 * annotations that contain boolean values.
 * </p>
 * <p>
 * This is a workaround using 'screenshots' of the checkboxes, since {@link TableViewer} is not able
 * to display SWT widgets when a cell is not edited. This concretely means that, when the cell is
 * not being edited (e.g., if no or another cell is edited), then the checkbox would simply
 * disappear and the default text-based label provider would be shown.
 * </p>
 * 
 * <p>
 * Since the screenshots are made during runtime, the resulting images fit the platform on which AF3
 * is run.
 * </p>
 * 
 * 
 * @author diewald
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating GREEN Hash: D480092E250130095120B89670E5D4DF
 */
public class EmulatedNativeCheckBoxLabelProvider extends AnnotationLabelProvider {

	/** Key for the 'checked' checkbox. */
	private static final String CHECKED_KEY = "CHECKED";

	/** Key for the 'unchecked' checkbox. */
	private static final String UNCHECKED_KEY = "UNCHECKED";

	/** Currently referenced row. */
	ViewerRow row;

	/** Constructs a new {@link EmulatedNativeCheckBoxLabelProvider}. */
	public EmulatedNativeCheckBoxLabelProvider(Class<? extends IAnnotatedSpecification> specClass,
			IAnnotationViewPart viewPart, String instanceKey, ColumnViewer viewer) {
		super(specClass, viewPart, instanceKey, null);

		if(getImageRegistry().getDescriptor(CHECKED_KEY) == null) {
			getImageRegistry().put(UNCHECKED_KEY, makeShot(viewer.getControl(), false));
			getImageRegistry().put(CHECKED_KEY, makeShot(viewer.getControl(), true));
		}
	}

	/** {@inheritDoc} */
	@Override
	public void update(final ViewerCell cell) {
		super.update(cell);

		if(cell != null && cell.getViewerRow() != null) {

			// On the first invocation, register mouse listener to
			// cell's control (i.e., the table) in order to remove the
			// icon when the actual widget is to be shown.
			//
			// This cannot be achieved using a FocusListener since it
			// does not provide a way to determine the cell which has been
			// clicked (source and widget of the event refer to the Table itself).
			if(row == null) {
				cell.getControl().addMouseListener(new MouseListener() {

					@Override
					public void mouseDoubleClick(MouseEvent e) {
						// Nothing to do
					}

					@Override
					public void mouseDown(MouseEvent e) {
						Point pt = new Point(e.x, e.y);

						// Registration of mouseListener ensures that row is non-null
						TableItem item = ((Table)row.getControl()).getItem(pt);

						// Hide image showing the icon emulating the native check box since the
						// user as clicked this cell and the check box widget is about to be
						// shown.
						//
						// Otherwise, both the icon and the widget would be visible in parallel.
						//
						// After the user has finished editing, the framework will call
						// getImage() in order to restore the icon emulating the native check
						// box.
						if(item != null) {

							// Source and widget of the event 'e' refer to the Table itself, so
							// we need to iterate over all columns of the current row and check the
							// coordinates to find the actual cell.
							for(int col = 0; col < ((Table)row.getControl()).getColumnCount(); col++) {
								Rectangle rect = item.getBounds(col);
								if(rect.contains(pt)) {
									row.setImage(col, null);
								}
							}
						}
					}

					@Override
					public void mouseUp(MouseEvent e) {
						// Nothing to do
					}
				});
			}

			// Later on, just update the referenced row.
			row = cell.getViewerRow();
		}
		super.update(cell);
	}

	/** Takes a 'screenshot' of the checkbox control an stores it in the image registry. */
	private Image makeShot(Control control, boolean type) {
		// Hopefully no platform uses exactly this color because we'll make
		// it transparent in the image.
		Color greenScreen = new Color(control.getDisplay(), 222, 223, 224);

		Shell shell = new Shell(control.getShell(), SWT.NO_TRIM);

		// otherwise we have a default gray color
		shell.setBackground(greenScreen);

		Button button = new Button(shell, SWT.CHECK);
		button.setBackground(greenScreen);
		button.setSelection(type);

		// otherwise an image is located in a corner
		button.setLocation(1, 1);
		Point bsize = button.computeSize(SWT.DEFAULT, SWT.DEFAULT);

		// otherwise an image is stretched by width
		bsize.x = Math.max(bsize.x - 1, bsize.y - 1);
		bsize.y = Math.max(bsize.x - 1, bsize.y - 1);
		button.setSize(bsize);
		shell.setSize(bsize);

		shell.open();
		GC gc = new GC(shell);
		Image image = new Image(control.getDisplay(), bsize.x, bsize.y);
		gc.copyArea(image, 0, 0);
		gc.dispose();
		shell.close();

		ImageData imageData = image.getImageData();
		imageData.transparentPixel = imageData.palette.getPixel(greenScreen.getRGB());

		return new Image(control.getDisplay(), imageData);
	}

	/** {@inheritDoc} */
	@Override
	public String getText(Object element) {
		return "";
	}

	/** {@inheritDoc} */
	@Override
	public Image getImage(Object element) {
		if(element instanceof AnnotationEntry && ((AnnotationEntry)element).canEdit(clazz)) {
			if(isChecked(element)) {
				return JFaceResources.getImageRegistry().get(CHECKED_KEY);
			}
			return JFaceResources.getImageRegistry().get(UNCHECKED_KEY);
		}
		return null;
	}

	/**
	 * Determines whether the boolean value is 'true' or 'false'. Used to determine the 'checked'
	 * status of the checkbox.
	 */
	protected boolean isChecked(Object element) {
		if(element instanceof AnnotationEntry) {
			AnnotationEntry annotationEntry = (AnnotationEntry)element;

			Object value = null;
			try {
				value = annotationEntry.getSpecificationValue(clazz, instanceKey);
			} catch(Exception e) {
				return false;
			}
			if(value != null) {
				// It is already known that the value is a boolean.
				return (Boolean)value;
			}
		}
		return false;
	}
}
