/*--------------------------------------------------------------------------+
$Id: DiagramEditorBase.java 13954 2015-06-03 09:07:23Z aravantinos $
|                                                                          |
| Copyright 2011 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.ui.editor;

import static org.eclipse.gef.SnapToGrid.PROPERTY_GRID_ENABLED;
import static org.eclipse.gef.SnapToGrid.PROPERTY_GRID_ORIGIN;
import static org.eclipse.gef.SnapToGrid.PROPERTY_GRID_SPACING;
import static org.eclipse.gef.SnapToGrid.PROPERTY_GRID_VISIBLE;
import static org.eclipse.gef.editparts.ZoomManager.FIT_ALL;
import static org.eclipse.gef.editparts.ZoomManager.FIT_HEIGHT;
import static org.eclipse.gef.editparts.ZoomManager.FIT_WIDTH;
import static org.fortiss.tooling.base.layout.DefaultLayoutConstants.DEFAULT_GRID_ORIGIN_X;
import static org.fortiss.tooling.base.layout.DefaultLayoutConstants.DEFAULT_GRID_ORIGIN_Y;
import static org.fortiss.tooling.base.layout.DefaultLayoutConstants.DEFAULT_GRID_SIZE;
import static org.fortiss.tooling.kernel.ui.util.EObjectSelectionUtils.getEObjectElements;
import static org.fortiss.tooling.kernel.ui.util.EObjectSelectionUtils.getFirstElement;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.conqat.ide.commons.ui.selection.SelectionUtils;
import org.eclipse.core.runtime.Platform;
import org.eclipse.draw2d.ColorConstants;
import org.eclipse.draw2d.IFigure;
import org.eclipse.draw2d.geometry.Dimension;
import org.eclipse.draw2d.geometry.Point;
import org.eclipse.emf.common.notify.Adapter;
import org.eclipse.emf.common.notify.Notification;
import org.eclipse.emf.common.util.TreeIterator;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.util.EContentAdapter;
import org.eclipse.gef.EditDomain;
import org.eclipse.gef.EditPart;
import org.eclipse.gef.EditPartFactory;
import org.eclipse.gef.EditPartListener;
import org.eclipse.gef.EditPartViewer;
import org.eclipse.gef.GraphicalViewer;
import org.eclipse.gef.MouseWheelHandler;
import org.eclipse.gef.MouseWheelZoomHandler;
import org.eclipse.gef.Request;
import org.eclipse.gef.RootEditPart;
import org.eclipse.gef.SnapToGeometry;
import org.eclipse.gef.commands.CommandStack;
import org.eclipse.gef.editparts.AbstractGraphicalEditPart;
import org.eclipse.gef.editparts.ScalableFreeformRootEditPart;
import org.eclipse.gef.editparts.ScalableRootEditPart;
import org.eclipse.gef.editparts.ZoomManager;
import org.eclipse.gef.editpolicies.AbstractEditPolicy;
import org.eclipse.gef.ui.actions.ActionRegistry;
import org.eclipse.gef.ui.actions.DeleteAction;
import org.eclipse.gef.ui.actions.PrintAction;
import org.eclipse.gef.ui.actions.SelectAllAction;
import org.eclipse.gef.ui.actions.SelectionAction;
import org.eclipse.gef.ui.actions.UpdateAction;
import org.eclipse.gef.ui.actions.ZoomComboContributionItem;
import org.eclipse.gef.ui.actions.ZoomInAction;
import org.eclipse.gef.ui.actions.ZoomOutAction;
import org.eclipse.jface.action.IAction;
import org.eclipse.jface.action.MenuManager;
import org.eclipse.jface.viewers.IPostSelectionProvider;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.jface.viewers.ISelectionChangedListener;
import org.eclipse.swt.SWT;
import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Menu;
import org.eclipse.ui.IActionBars;
import org.eclipse.ui.IEditorInput;
import org.eclipse.ui.IEditorSite;
import org.eclipse.ui.ISelectionListener;
import org.eclipse.ui.IWorkbenchPart;
import org.eclipse.ui.PartInitException;
import org.fortiss.tooling.base.model.element.IConnection;
import org.fortiss.tooling.base.model.element.IConnector;
import org.fortiss.tooling.base.model.element.IHierarchicElement;
import org.fortiss.tooling.base.ui.dnd.gef.ElementCompositionDropTargetListener;
import org.fortiss.tooling.base.ui.editpart.ConnectorEditPartBase;
import org.fortiss.tooling.base.ui.editpart.DiagramEditPartBase;
import org.fortiss.tooling.base.ui.editpart.ExtendedLayerRootEditPart;
import org.fortiss.tooling.base.ui.layout.IDiagramLayoutConfiguration;
import org.fortiss.tooling.kernel.model.ILibraryElementReference;
import org.fortiss.tooling.kernel.service.ILibraryService;
import org.fortiss.tooling.kernel.ui.extension.base.EObjectActionBase;
import org.fortiss.tooling.kernel.ui.extension.data.ContextMenuContextProvider;
import org.fortiss.tooling.kernel.ui.service.IActionService;
import org.fortiss.tooling.kernel.ui.service.IContextMenuService;
import org.fortiss.tooling.kernel.ui.service.IEditPartFactoryService;

/**
 * <p>
 * GEF editor base implementation for graphical viewer editors.
 * </p>
 * <p>
 * It provides {@link EditDomain} and thus a {@link CommandStack}.
 * </p>
 * 
 * @author hoelzl
 * @author $Author: aravantinos $
 * @version $Rev: 13954 $
 * @ConQAT.Rating YELLOW Hash: C96404EE24E8C30B5A2AABFB1342F04A
 */
abstract class CommonDiagramEditorBase<T extends EObject> extends GEFEditorBase<T> implements
		IPostSelectionProvider, ContextMenuContextProvider {

	/** Stores the viewer. */
	private GraphicalViewer viewer;

	/** Stores the edit part factory */
	private EditPartFactory editPartFactory;

	/** Actions which need update upon selection changes. */
	private final List<String> selectionActions = new ArrayList<String>();

	/** Actions which need update upon changes of properties. */
	private final List<String> propertyActions = new ArrayList<String>();

	/** Stores the zoom manager. */
	private ZoomManager zoomManager;

	/** Stores the zoom contribution item. */
	private ZoomComboContributionItem zoomContribution;

	/** A container for editor actions. */
	private final ActionRegistry actionRegistry = new ActionRegistry();

	/**
	 * Number of times the hint was displayed until we assume the user knows how to draw a
	 * connection.
	 */
	private static int numberOfHintDisplays = 1;

	/** Returns numberOfHintDisplays *and decrements it*. */
	private static boolean shouldDisplayHint() {
		return numberOfHintDisplays-- >= 0;
	}

	/** Constructor. */
	public CommonDiagramEditorBase(GraphicalViewer viewer, EditPartFactory factory) {
		this.viewer = viewer;
		this.editPartFactory = factory;
	}

	/**
	 * Adapter to watch for added {@link IConnection}s (used to update {@link #numberOfHintDisplays}
	 * ).
	 */
	private static final Adapter addConnectionAdapter = new EContentAdapter() {
		@Override
		public void notifyChanged(Notification notification) {

			super.notifyChanged(notification);
			int eventType = notification.getEventType();

			if(eventType == Notification.REMOVING_ADAPTER) {
				return;
			}

			// Only count the notification by the IHierarchicElement owning the new IConnection
			Object newValue = notification.getNewValue();
			if(eventType == Notification.ADD && newValue instanceof IConnection &&
					notification.getNotifier() == ((EObject)newValue).eContainer()) {
				if(!shouldDisplayHint()) {
					// Hints will no longer be displayed: remove adapter
					EObject rootObj = (EObject)notification.getNotifier();
					while(rootObj.eContainer() != null) {
						rootObj = rootObj.eContainer();
					}
					rootObj.eAdapters().remove(addConnectionAdapter);
					for(TreeIterator<EObject> iter = rootObj.eAllContents(); iter.hasNext();) {
						EObject obj = iter.next();
						obj.eAdapters().remove(addConnectionAdapter);
					}
				}
			}
		}
	};

	/** Selection listener used to update selection actions. */
	private final ISelectionListener selectionListener = new ISelectionListener() {

		/** {@inheritDoc} */
		@Override
		public void selectionChanged(IWorkbenchPart part, ISelection selection) {
			EditPart ep = SelectionUtils.checkAndPickFirstSafe(selection, EditPart.class);
			if(ep != null) {
				Control control = ep.getViewer().getControl();
				Object model = ep.getModel();

				String hint = "";
				if(shouldDisplayHint()) {
					String key = Platform.getOS().equals(Platform.OS_LINUX) ? "Ctrl" : "Alt";
					String commonStr = " by pressing " + key + " and dragging";
					if(model instanceof IConnector) {
						hint = "Link two connectors" + commonStr + "from source to target";
					} else if(model instanceof IHierarchicElement) {
						hint = "Create a connection" + commonStr;
					}
					hint +=
							" (select anything on the diagram to prevent further display of this message).";
				}
				control.setToolTipText(hint);
			}
			updateActions(selectionActions);
		}
	};

	/** {@inheritDoc} */
	@Override
	public void init(IEditorSite site, IEditorInput input) throws PartInitException {
		super.init(site, input);

		getSite().getWorkbenchWindow().getSelectionService()
				.addSelectionListener(selectionListener);

		// Install adapter to top-most object to watch for addition of IConnections.
		if(shouldDisplayHint()) {
			EObject model = getEditedObject();
			EObject obj = model;
			while(obj != null) {
				obj = obj.eContainer();
			}

			if(!model.eAdapters().contains(addConnectionAdapter)) {
				model.eAdapters().add(addConnectionAdapter);
			}
		}
		viewer.setKeyHandler(new DiagramKeyHandler(viewer));
	}

	/** {@inheritDoc} */
	@Override
	public EObject getSelectedModelElement() {
		if(getSite().getSelectionProvider() != null) {
			return getFirstElement(getSite().getSelectionProvider().getSelection());
		}
		return null;
	}

	/** {@inheritDoc} */
	@Override
	public List<EObject> getSelectedModelElementList() {
		if(getSite().getSelectionProvider() != null) {
			return getEObjectElements(getSite().getSelectionProvider().getSelection());
		}
		return null;
	}

	/** Creates the context menu. */
	private void createContextMenu() {
		MenuManager menuManager = IContextMenuService.getInstance().createDefaultContextMenu(this);

		Menu contextMenu = menuManager.createContextMenu(viewer.getControl());
		viewer.getControl().setMenu(contextMenu);

		getSite().registerContextMenu(menuManager, viewer);
	}

	/** {@inheritDoc} */
	@Override
	public void createPartControl(Composite parent) {
		viewer.setRootEditPart(new ExtendedLayerRootEditPart());
		viewer.createControl(parent);

		getEditDomain().addViewer(viewer);

		viewer.setEditPartFactory(editPartFactory);
		viewer.setContents(getViewerContentObject());

		createContextMenu();

		getSite().setSelectionProvider(this);

		createActions();
		updateActions(propertyActions);

		createZoomSupport();

		// drag and drop
		viewer.addDropTargetListener(new ElementCompositionDropTargetListener(viewer));

		viewer.getControl().setBackground(parent.getBackground());

		if(enableSnapToGrid()) {
			viewer.setProperty(SnapToGeometry.PROPERTY_SNAP_ENABLED, true);
			viewer.setProperty(PROPERTY_GRID_ORIGIN, new Point(DEFAULT_GRID_ORIGIN_X,
					DEFAULT_GRID_ORIGIN_Y));

			Dimension gridSpacing = new Dimension(DEFAULT_GRID_SIZE, DEFAULT_GRID_SIZE);
			viewer.setProperty(PROPERTY_GRID_SPACING, gridSpacing);
			viewer.setProperty(PROPERTY_GRID_ENABLED, true);
			viewer.setProperty(PROPERTY_GRID_VISIBLE, true);
		}
		// this is needed by direct connection feature. the root edit part
		// becomes transparent when GEF selects a tool on button down event.
		// we have to relocate the call to the underlying root's content
		((ScalableRootEditPart)viewer.getRootEditPart()).installEditPolicy("Test",
				new AbstractEditPolicy() {
					@Override
					public EditPart getTargetEditPart(Request request) {
						return getHost().getRoot().getContents().getTargetEditPart(request);
					}
				});

		if(ILibraryService.getInstance().isLibraryElementShadow(editedObject) ||
				(editedObject instanceof ILibraryElementReference)) {
			((ScalableRootEditPart)viewer.getRootEditPart()).deactivate();
		}

		// Ensure that the editor gets the focus when a child edit part is added or removed in order
		// to fix undo/redo (at least as long as either a DiagramEditorBase or the NavigatorViewPart
		// are used to manipulate the model)
		viewer.getRootEditPart().getContents().addEditPartListener(new EditPartListener() {
			@Override
			public void selectedStateChanged(EditPart editpart) {
				// Nothing
			}

			@Override
			public void removingChild(EditPart child, int index) {
				viewer.getControl().setFocus();
			}

			@Override
			public void partDeactivated(EditPart editpart) {
				// Nothing
			}

			@Override
			public void partActivated(EditPart editpart) {
				// Nothing
			}

			@Override
			public void childAdded(EditPart child, int index) {
				viewer.getControl().setFocus();
			}
		});
	}

	/**
	 * Returns the root object used as content root. Sub-classes may override to
	 * provide a different content root object. The default implementation uses
	 * {@link GEFEditorBase#getEditedObject()}
	 */
	protected Object getViewerContentObject() {
		return getEditedObject();
	}

	/**
	 * Returns whether snap-to-grid should be enabled. The default is <code>true</code>. IMPORTANT:
	 * The return value provided here must match
	 * the value of {@link IDiagramLayoutConfiguration#useGrid()} provided by
	 * the {@link DiagramEditPartBase} of the diagram element.
	 */
	// TODO (FH): https://af3.fortiss.org/issues/376
	// this should also use the configuration automatically
	// making the IMPORTANT note irrelevant
	protected boolean enableSnapToGrid() {
		return true;
	}

	/** Installs zoom support for this graphical viewer. */
	private void createZoomSupport() {
		RootEditPart rootPart = viewer.getRootEditPart();
		if(rootPart instanceof ScalableFreeformRootEditPart) {
			zoomManager = ((ScalableFreeformRootEditPart)rootPart).getZoomManager();
		} else if(viewer.getRootEditPart() instanceof ScalableRootEditPart) {
			zoomManager = ((ScalableRootEditPart)rootPart).getZoomManager();
		}

		if(zoomManager != null) {
			zoomManager.setZoomLevelContributions(getZoomLevelContributions());
			registerZoomAction();
		}
		zoomContribution = new ZoomComboContributionItem(getSite().getPage());
	}

	/** Registers and returns the id of the given action. */
	protected final String registerAction(IAction action) {
		actionRegistry.registerAction(action);
		return action.getId();
	}

	/** Returns the zoom manager. */
	protected ZoomManager getZoomManager() {
		return zoomManager;
	}

	/** Registers zoom actions. */
	private void registerZoomAction() {
		registerAction(new ZoomInAction(zoomManager));
		registerAction(new ZoomOutAction(zoomManager));

		viewer.setProperty(MouseWheelHandler.KeyGenerator.getKey(SWT.MOD1),
				MouseWheelZoomHandler.SINGLETON);
	}

	/** {@inheritDoc} */
	@SuppressWarnings("rawtypes")
	@Override
	public Object getAdapter(Class adapter) {
		// This is called, e.g., from the PrintAction
		if(adapter == GraphicalViewer.class) {
			return viewer;
		}
		// This adapter is required for the ZoomComboContributionItem
		if(adapter == ZoomManager.class) {
			return getZoomManager();
		}
		return super.getAdapter(adapter);
	}

	/**
	 * Creates the {@link EditPartFactory} which is used for this editor. The
	 * default implementation uses the {@link EditPartFactory}.
	 * TODO: remove
	 */
	protected EditPartFactory createEditPartFactory() {
		return IEditPartFactoryService.getInstance();
	}

	/**
	 * Returns the zoom level contributions. The default implementation allows
	 * {@link ZoomManager#FIT_ALL}, {@link ZoomManager#FIT_WIDTH}, and
	 * {@link ZoomManager#FIT_HEIGHT}.
	 */
	protected List<String> getZoomLevelContributions() {
		return defaultZoomLevelContribution;
	}

	/**
	 * Default zoom level contributions.
	 */
	private static final List<String> defaultZoomLevelContribution = Arrays.asList(new String[] {
			FIT_ALL, FIT_WIDTH, FIT_HEIGHT});

	/** Registers global actions (including zoom action). */
	@Override
	public void registerGlobalActions(IActionBars bars) {
		bars.getToolBarManager().add(zoomContribution);
	}

	/** Unregisters global actions (including zoom action). */
	@Override
	public void unregisterGlobalActions(IActionBars bars) {
		bars.getToolBarManager().remove(zoomContribution);
	}

	/**
	 * Registers the select all action. Sub-classes may override to prevent the
	 * action registering.
	 */
	protected void registerSelectAllAction() {
		registerAction(new SelectAllAction(this));
	}

	/**
	 * Registers the print action. Sub-classes may override to prevent the
	 * action registering.
	 */
	protected void registerPrintAction() {
		registerAction(new PrintAction(this));
	}

	/**
	 * Register any other action defined by sub-classes. Sub-classes should
	 * override to implement.
	 */
	protected void registerOtherActions() {
		// do nothing by default
	}

	/**
	 * Registers the copy, cut, and past action. Sub-classes may override to
	 * prevent the action registering.
	 */
	protected void registerEditActions() {
		registerSelectionAction(new DeleteAction((IWorkbenchPart)this));
	}

	/** Registers and returns the id of the given selection action. */
	protected final String registerSelectionAction(SelectionAction action) {
		selectionActions.add(registerAction(action));
		action.setSelectionProvider(viewer);
		return action.getId();
	}

	/** Creates actions for this editor. */
	private void createActions() {
		registerSelectAllAction();
		registerPrintAction();
		registerEditActions();
		registerOtherActions();
	}

	/** Updates a list of actions. */
	@SuppressWarnings({"rawtypes", "unchecked"})
	private void updateActions(List<String> actionIds) {
		EObject element = getSelectedModelElement();
		for(String id : actionIds) {
			IAction action = actionRegistry.getAction(id);
			if(action instanceof UpdateAction) {
				((UpdateAction)action).update();
			} else if(action instanceof EObjectActionBase) {
				((EObjectActionBase)action).refresh(element);
			}
		}
		IActionService.getInstance().refresh();
	}

	/** {@inheritDoc} */
	@Override
	public void setFocus() {
		viewer.getControl().setFocus();
	}

	/** {@inheritDoc} */
	@Override
	protected void firePropertyChange(int property) {
		super.firePropertyChange(property);
		updateActions(propertyActions);
	}

	// The following methods are necessary in order to avoid Eclipse Helios SR2
	// bug #339360

	/** {@inheritDoc} */
	@Override
	public void addPostSelectionChangedListener(ISelectionChangedListener listener) {
		viewer.addSelectionChangedListener(listener);
	}

	/** {@inheritDoc} */
	@Override
	public void removePostSelectionChangedListener(ISelectionChangedListener listener) {
		viewer.removeSelectionChangedListener(listener);
	}

	/** {@inheritDoc} */
	@Override
	public void addSelectionChangedListener(ISelectionChangedListener listener) {
		viewer.addSelectionChangedListener(listener);
	}

	/** {@inheritDoc} */
	@Override
	public ISelection getSelection() {
		return viewer.getSelection();
	}

	/** {@inheritDoc} */
	@Override
	public void removeSelectionChangedListener(ISelectionChangedListener listener) {
		viewer.removeSelectionChangedListener(listener);
	}

	/** {@inheritDoc} */
	@Override
	public void setSelection(ISelection selection) {
		viewer.setSelection(selection);
	}

	/** {@inheritDoc} */
	@Override
	public void setHighlight(EObject element, boolean highlighted) {
		EditPart ep = findEditPart(viewer.getRootEditPart(), element);
		if(ep instanceof AbstractGraphicalEditPart) {
			IFigure fig = ((AbstractGraphicalEditPart)ep).getFigure();
			setHighlight(fig, highlighted);
		}
	}

	/**
	 * Highlights a figure by setting its color to be RED. This implementation
	 * can be changed by more specific editors.
	 */
	protected void setHighlight(IFigure fig, boolean highlighted) {
		if(highlighted) {
			if(!highlightedFigures2NormalFgColor.containsKey(fig)) {
				highlightedFigures2NormalFgColor.put(fig, fig.getForegroundColor());
				fig.setForegroundColor(ColorConstants.red);
				highlightedFigures2NormalBgColor.put(fig, fig.getBackgroundColor());
				fig.setBackgroundColor(ColorConstants.red);
			}
		} else {
			Color normalColor = highlightedFigures2NormalFgColor.get(fig);
			fig.setForegroundColor(normalColor);
			normalColor = highlightedFigures2NormalBgColor.get(fig);
			fig.setBackgroundColor(normalColor);
			highlightedFigures2NormalFgColor.remove(fig);
		}
	}

	/** A map from figures that are currently highlighted to their normal foreground color. */
	private HashMap<IFigure, Color> highlightedFigures2NormalFgColor =
			new HashMap<IFigure, Color>();

	/** A map from figures that are currently highlighted to their normal foreground color. */
	private HashMap<IFigure, Color> highlightedFigures2NormalBgColor =
			new HashMap<IFigure, Color>();

	/** Returns the edit part corresponding to the EObject or null if it cannot be found. */
	@SuppressWarnings("unchecked")
	private EditPart findEditPart(EditPart container, EObject element) {
		if(container.getModel().equals(element)) {
			return container;
		}

		List<EditPart> subEditParts = new ArrayList<EditPart>();
		if(container.getChildren() != null) {
			subEditParts.addAll(container.getChildren());
		}
		if(container instanceof ConnectorEditPartBase) {
			ConnectorEditPartBase<?> connector = (ConnectorEditPartBase<?>)container;
			if(connector.getSourceConnections() != null) {
				subEditParts.addAll(connector.getSourceConnections());
			}
		}

		for(EditPart child : subEditParts) {
			EditPart foundEditPart = findEditPart(child, element);
			if(foundEditPart != null) {
				return foundEditPart;
			}
		}

		return null;
	}

	/** {@inheritDoc} */
	@Override
	public void clearAllHighlights() {
		Set<IFigure> figs = new HashSet<IFigure>(highlightedFigures2NormalFgColor.keySet());

		for(IFigure figure : figs) {
			setHighlight(figure, false);
		}
	}

	/** Returns the diagram editor's viewer. */
	public EditPartViewer getViewer() {
		return viewer;
	}
}
