/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2011 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.ui.utils;

import static org.fortiss.tooling.base.utils.LayoutModelElementFactory.createDimension;
import static org.fortiss.tooling.base.utils.LayoutModelElementFactory.createPoint;

import org.eclipse.draw2d.geometry.Dimension;
import org.eclipse.draw2d.geometry.Point;
import org.eclipse.gef.RootEditPart;
import org.eclipse.gef.editparts.ScalableFreeformRootEditPart;
import org.eclipse.gef.editparts.ScalableRootEditPart;
import org.eclipse.gef.requests.ChangeBoundsRequest;
import org.fortiss.tooling.base.utils.ZoomUtils;

/**
 * Utility code to support zooming.
 * 
 * @author hummel
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating GREEN Hash: 663B595CF51284798F00160254E64C68
 */
public class ZoomUIUtils {

	/** Extracts the current zoom level from the given root edit part. */
	public static double determineZoom(RootEditPart rootEditPart) {
		if(rootEditPart instanceof ScalableRootEditPart) {
			return ((ScalableRootEditPart)rootEditPart).getZoomManager().getZoom();
		}
		if(rootEditPart instanceof ScalableFreeformRootEditPart) {
			return ((ScalableFreeformRootEditPart)rootEditPart).getZoomManager().getZoom();
		}
		return 1;
	}

	/**
	 * Fixes a {@link ChangeBoundsRequest} for the current zoom level. Both the
	 * move and size delta are changed.
	 */
	public static void fixChangeBoundRequest(ChangeBoundsRequest request, double zoom) {
		if(zoom <= 0 || zoom == 1) {
			return;
		}
		request.setMoveDelta(screenToInternalPoint(request.getMoveDelta(), zoom));
		request.setSizeDelta(screenToInternalDimension(request.getSizeDelta(), zoom));
	}

	/**
	 * Converts a point expressed in screen coordinates, into a point in internal coordinates,
	 * taking the zoom into account.
	 */
	public static Point screenToInternalPoint(Point pos, double zoom) {
		org.fortiss.tooling.base.model.layout.Point modelPoint =
				createPoint(pos.x(), pos.y(), pos.toString());
		modelPoint = ZoomUtils.screenToInternalPoint(modelPoint, zoom);
		pos.setLocation(modelPoint.getX(), modelPoint.getY());
		return pos;
	}

	/**
	 * Converts a point expressed in internal coordinates, into a point in screen coordinates,
	 * taking the zoom into account.
	 */
	public static Point internalToScreenPoint(Point pos, double zoom) {
		org.fortiss.tooling.base.model.layout.Point modelPoint =
				createPoint(pos.x(), pos.y(), pos.toString());
		modelPoint = ZoomUtils.internalToScreenPoint(modelPoint, zoom);
		pos.setLocation(modelPoint.getX(), modelPoint.getY());
		return pos;
	}

	/**
	 * Converts a point expressed in screen coordinates, into a point in internal coordinates,
	 * taking the zoom into account.
	 */
	public static Dimension screenToInternalDimension(Dimension dim, double zoom) {
		org.fortiss.tooling.base.model.layout.Dimension modelDim =
				createDimension(dim.width(), dim.height(), dim.toString());
		modelDim = ZoomUtils.screenToInternalDimension(modelDim, zoom);
		dim.setSize(modelDim.getWidth(), modelDim.getHeight());
		return dim;
	}

	/**
	 * Converts a point expressed in internal coordinates, into a point in screen coordinates,
	 * taking the zoom into account.
	 */
	public static Dimension internalToScreenDimension(Dimension dim, double zoom) {
		org.fortiss.tooling.base.model.layout.Dimension modelDim =
				createDimension(dim.width(), dim.height(), dim.toString());
		modelDim = ZoomUtils.internalToScreenDimension(modelDim, zoom);
		dim.setSize(modelDim.getWidth(), modelDim.getHeight());
		return dim;
	}
}
