/*--------------------------------------------------------------------------+
$Id: codetemplates.xml 1 2011-01-01 00:00:01Z hoelzl $
|                                                                          |
| Copyright 2011 fortiss GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.compose;

import static org.eclipse.emf.ecore.util.EcoreUtil.delete;

import java.util.ArrayList;

import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.util.EcoreUtil;
import org.fortiss.tooling.base.model.element.IConnection;
import org.fortiss.tooling.base.model.element.IConnector;
import org.fortiss.tooling.base.model.element.IHierarchicElement;
import org.fortiss.tooling.base.model.element.IHierarchicElementContainer;
import org.fortiss.tooling.kernel.service.IConnectionCompositorService;
import org.fortiss.tooling.kernel.service.IElementCompositorService;

/**
 * Base implementation of a compositor for {@link IHierarchicElementContainer}s
 * focusing on the remove mechanism.
 * 
 * @author doebber
 * @author $Author: hoelzl $
 * @version $Rev: 18709 $
 * @ConQAT.Rating YELLOW Hash: D2D82D649AC5A637CDDFE276F604F388
 */
public abstract class HierarchicElementCompositorBase<HE extends IHierarchicElementContainer>
		extends ModelElementCompositorBase<HE> {
	/**
	 * Decomposition that takes care of the whole model subtree. Subclasses
	 * implement {@code #decompose*****()} instead.
	 */
	@Override
	public final boolean decompose(EObject contained) {
		return iterateDecompose(contained) && decomposeSpecific(contained);
	}

	/** Recursive traversal of all model elements in subtree. */
	@Override
	protected boolean iterateDecompose(EObject contained) {
		super.iterateDecompose(contained);
		if(contained instanceof IHierarchicElementContainer) {
			if(!iterateDecomposeSubelements((IHierarchicElementContainer)contained)) {
				return false;
			}
		}
		if(contained instanceof IHierarchicElement) {
			IHierarchicElement he = (IHierarchicElement)contained;

			if(!iterateDecomposeConnections(he.getConnections())) {
				return false;
			}

			if(!iterateDecomposeConnectors((IHierarchicElement)contained)) {
				return false;
			}
		}
		if(contained instanceof IConnector) {
			IConnector connector = (IConnector)contained;
			if(!iterateDecomposeConnections(connector.getIncoming())) {
				return false;
			}
			if(!iterateDecomposeConnections(connector.getOutgoing())) {
				return false;
			}
		}
		return decomposeSpecific(contained);
	}

	/**
	 * Iterates over sub elements of given {@link IHierarchicElementContainer} and performs
	 * decomposition.
	 */
	private boolean iterateDecomposeSubelements(IHierarchicElementContainer contained) {
		ArrayList<IHierarchicElement> list =
				new ArrayList<IHierarchicElement>(contained.getContainedElements());
		for(IHierarchicElement subelement : list) {
			if(!decomposeSubelement(subelement)) {
				return false;
			}
		}
		return true;
	}

	/**
	 * Iterates over list of connections and runs disconnection or decomposition
	 * of sub elements respectively.
	 */
	private boolean iterateDecomposeConnections(EList<IConnection> connectionsList) {
		ArrayList<IConnection> list = new ArrayList<IConnection>(connectionsList);
		for(IConnection conn : list) {
			if(!decomposeConnection(conn)) {
				return false;
			}
		}
		return true;
	}

	/**
	 * Iterates over connectors of given {@link IHierarchicElement} and runs
	 * decomposition.
	 */
	private boolean iterateDecomposeConnectors(IHierarchicElement contained) {
		ArrayList<IConnector> list = new ArrayList<IConnector>(contained.getConnectors());
		for(IConnector c : list) {
			if(!decomposeConnector(c)) {
				return false;
			}
		}
		return true;
	}

	/**
	 * Base implementation returns compositors' decomposition by default and
	 * invokes {@link EcoreUtil#delete(EObject, boolean)} otherwise. Subclasses may override.
	 */
	protected boolean decomposeSubelement(IHierarchicElement element) {
		if(IElementCompositorService.getInstance().canDecompose(element)) {
			return IElementCompositorService.getInstance().decompose(element);
		}
		delete(element, true);
		return true;
	}

	/**
	 * Base implementation returns compositors' decomposition by default and
	 * invokes {@link EcoreUtil#delete(EObject, boolean)} otherwise. Subclasses may override.
	 */
	protected boolean decomposeConnector(IConnector element) {
		ArrayList<IConnection> list = new ArrayList<IConnection>(element.getIncoming());
		list.addAll(element.getOutgoing());
		for(IConnection conn : list) {
			if(!IConnectionCompositorService.getInstance().disconnect(conn)) {
				return false;
			}
		}

		if(IElementCompositorService.getInstance().canDecompose(element)) {
			return IElementCompositorService.getInstance().decompose(element);
		}

		EcoreUtil.delete(element, true);
		return true;
	}

	/**
	 * Base implementation returns compositors' decomposition by default and
	 * invokes {@link EcoreUtil#delete(EObject, boolean)} otherwise. Subclasses may override.
	 */
	protected boolean decomposeConnection(IConnection conn) {
		return IConnectionCompositorService.getInstance().disconnect(conn);
	}
}
