/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2014 fortiss GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.utils;

import static org.fortiss.tooling.kernel.utils.EcoreUtils.getChildrenWithType;
import static org.fortiss.tooling.kernel.utils.EcoreUtils.getEClassForClass;
import static org.fortiss.tooling.kernel.utils.EcoreUtils.getParentsWithType;
import static org.fortiss.tooling.kernel.utils.EcoreUtils.pickFirstInstanceOf;

import java.util.List;

import org.eclipse.emf.common.util.BasicEList;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.util.EcoreUtil;
import org.fortiss.tooling.base.annotation.IAnnotationValueService;
import org.fortiss.tooling.base.compose.HierarchicElementCompositorBase;
import org.fortiss.tooling.base.model.element.IAnnotatedSpecification;
import org.fortiss.tooling.base.model.element.IModelElement;
import org.fortiss.tooling.kernel.extension.data.ITopLevelElement;
import org.fortiss.tooling.kernel.service.IPersistencyService;

/**
 * Utility methods for dealing with {@link IAnnotatedSpecification}s.
 * 
 * @author diewald
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating YELLOW Hash: CC8B6E824145972B0C6DCE2F8CA2B19E
 */
public class AnnotationUtils {
	/**
	 * Returns the annotation of the particular {@code clazz} from a given {@code modelElement}.
	 */
	private static <T extends IAnnotatedSpecification> T getAnnotationInternal(
			final IModelElement modelElement, Class<T> clazz) {
		return pickFirstInstanceOf(clazz, modelElement.getSpecifications());
	}

	/**
	 * Creates an annotation of a particular {@code eClass} and adds it the specification list of
	 * the {@code modelElement}.
	 */
	@SuppressWarnings("unchecked")
	private static <T extends IAnnotatedSpecification> void createAnnotationInternal(
			final IModelElement modelElement, Class<T> clazz) {

		final EClass eClass = getEClassForClass(clazz);

		if(eClass == null) {
			return;
		}

		final T newAnnotation = (T)EcoreUtil.create(eClass);
		modelElement.getSpecifications().add(newAnnotation);
	}

	/**
	 * Recursively instantiates the annotations for a given {@link IModelElement} and its offspring.
	 * This is useful to ensure that all annotations are available for programmatically created
	 * models. While the {@link HierarchicElementCompositorBase} could be used to ensure the
	 * instantiation of annotations in model-transformations, too, this is not always possible (or
	 * convenient).
	 * 
	 * @param element
	 *            {@link IModelElement} for which annotations should be instantiated recursively.
	 */
	public static void instantiateAnnotationsRecursive(IModelElement element) {
		IAnnotationValueService.getInstance().instantiateAnnotations(element);
		for(IModelElement child : getChildrenWithType(element, IModelElement.class)) {
			IAnnotationValueService.getInstance().instantiateAnnotations(child);
		}
	}

	/**
	 * <p>
	 * Returns the annotation of a given {@code Class<T>} from a given {@code modelElement}. If the
	 * annotation does not exist yet, it is created and added to the specification list of the
	 * {@code modelElement}.
	 * </p>
	 * 
	 * <p>
	 * <b>Note: </b>The access to the model is wrapped into a command in order to allow for
	 * transparent creation of the annotation object from other contexts (e.g., GUI). If this is not
	 * desirable, use the {@link #getAnnotation(IModelElement, Class, boolean)} variant of this
	 * method.
	 * </p>
	 */
	public static <T extends IAnnotatedSpecification> T getAnnotation(IModelElement modelElement,
			Class<T> clazz) {
		return getAnnotation(modelElement, clazz, true);
	}

	/**
	 * Returns the annotation of a given {@code Class<T>} from a given {@code modelElement}. If the
	 * annotation does not exist yet, it is created and added to the specification list of the
	 * {@code modelElement}. Depending on the caller's choice, the model modification is wrapped
	 * into a command.
	 */
	public static <T extends IAnnotatedSpecification> T getAnnotation(
			final IModelElement modelElement, final Class<T> clazz, boolean wrapIntoCommand) {

		T annotation = getAnnotationInternal(modelElement, clazz);

		if(annotation != null) {
			return annotation;
		}

		if(wrapIntoCommand) {
			ITopLevelElement modelContext =
					IPersistencyService.getInstance().getTopLevelElementFor(modelElement);
			modelContext.runAsCommand(() -> {
				// TODO (see #2228 / #2208) Keep this code that ensures that non-existing
				// annotations are lazily instantiated until #2228 is resolved. It should be
				// noted that each time an annotation is constructed from this context, an
				// undo/redo command will be created that is not comprehensible for the user
				// (see #2208).
					createAnnotationInternal(modelElement, clazz);
				});
		} else {
			// Create if the creation does not need to be wrapped into a command.
			createAnnotationInternal(modelElement, clazz);
		}

		// Return new annotation (be it created in a command, or directly). This will return null in
		// case the creation of the new annotation was not successful.
		return getAnnotationInternal(modelElement, clazz);
	}

	/**
	 * Returns the {@link IAnnotatedSpecification}s matching the type 'T' of
	 * the parents of the given {@link IModelElement}.
	 * 
	 * @param startElement
	 *            The {@link IModelElement} to start the search from.
	 * @param clazz
	 *            The {@link IAnnotatedSpecification}s to be searched for.
	 * @return {@link EList} of found annotations.
	 */
	public static <T extends IAnnotatedSpecification> EList<T> getParentAnnotationsWithType(
			IModelElement startElement, Class<T> clazz) {
		EList<T> parentAnnotationList = new BasicEList<T>();
		List<IModelElement> parentsElements = getParentsWithType(startElement, IModelElement.class);

		for(IModelElement parentElement : parentsElements) {
			T annotation = pickFirstInstanceOf(clazz, parentElement.getSpecifications());
			parentAnnotationList.add(annotation);
		}
		return parentAnnotationList;
	}
}
