/*--------------------------------------------------------------------------+
$Id: codetemplates.xml 1 2011-01-01 00:00:01Z hoelzl $
|                                                                          |
| Copyright 2011 fortiss GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.utils;

import java.util.ArrayList;
import java.util.Collection;
import java.util.LinkedList;
import java.util.List;

import org.fortiss.tooling.base.model.element.IConnection;
import org.fortiss.tooling.base.model.element.IConnector;
import org.fortiss.tooling.base.model.element.IHiddenSpecification;
import org.fortiss.tooling.base.model.element.IHierarchicElement;
import org.fortiss.tooling.base.model.element.IHierarchicElementContainer;
import org.fortiss.tooling.base.model.element.IModelElement;
import org.fortiss.tooling.base.model.element.IModelElementReference;
import org.fortiss.tooling.base.model.element.IModelElementSpecification;

/**
 * Utility class for model elements.
 * 
 * @author trachtenherz
 * @author mou
 * @author hoelzl
 * @author $Author: hoelzl $
 * @version $Rev: 18709 $
 * @ConQAT.Rating YELLOW Hash: 58521C469226B471D914409DD11925CD
 */
public class BaseModelElementUtils {

	/**
	 * Returns a list of specifications from {@code specifications} that are not
	 * subclass of {@link IHiddenSpecification}.
	 * 
	 * @param specifications
	 *            the list of specification to be filtered
	 * 
	 * @return {@link List} of {@link IModelElementSpecification} from {@code specifications} that
	 *         are not subclass of {@link IHiddenSpecification}.
	 */
	public static List<IModelElementSpecification> nonHiddenSpecifications(
			Collection<IModelElementSpecification> specifications) {
		List<IModelElementSpecification> res = new LinkedList<IModelElementSpecification>();
		for(IModelElementSpecification s : specifications) {
			if(!(s instanceof IHiddenSpecification)) {
				res.add(s);
			}
		}
		return res;
	}

	/**
	 * Return the index of the given {@link IHierarchicElement} in its parent
	 * {@link IHierarchicElementContainer}.
	 * 
	 * @param element
	 *            the hierarchic element
	 * 
	 * @return Index of the given {@link IHierarchicElement} in its parent
	 *         {@link IHierarchicElementContainer}.
	 */
	public static int getIndex(IHierarchicElement element) {
		IHierarchicElementContainer container = element.getContainer();
		if(container != null) {
			return container.getContainedElements().indexOf(element);
		}
		return -1;
	}

	/**
	 * Removes {@link IModelElementSpecification}s of a {@link IModelElement} that represent a
	 * {@link IModelElementReference}s to another model element.
	 * 
	 * @param modelElement
	 *            the {@link IModelElement} of which to remove the
	 *            {@link IModelElementSpecification}s
	 *            representing {@link IModelElementReference}s.
	 * 
	 * @return Whether the {@link IModelElementSpecification}s representing
	 *         {@link IModelElementReference} have been removed, i.e. if the model has been
	 *         modified.
	 */
	public static boolean removeModelElementReferenceSpecifications(IModelElement modelElement) {
		List<IModelElementSpecification> modelElementReferenceSpecifications =
				findModelElementReferenceSpecifications((modelElement).getSpecifications());

		boolean modelElementReferenceSpecificationsRemoved =
				!modelElementReferenceSpecifications.isEmpty();

		modelElement.getSpecifications().removeAll(modelElementReferenceSpecifications);

		return modelElementReferenceSpecificationsRemoved;
	}

	/**
	 * Finds all {@link IModelElementSpecification}s of a {@link IModelElement} that represent a
	 * {@link IModelElementReference}s to another model element.
	 */
	private static List<IModelElementSpecification> findModelElementReferenceSpecifications(
			Collection<IModelElementSpecification> specifications) {
		List<IModelElementSpecification> result = new ArrayList<IModelElementSpecification>();
		for(IModelElementSpecification specification : specifications) {
			if(specification instanceof IModelElementReference)
				result.add(specification);
		}
		return result;
	}

	/**
	 * For a given {@link IConnector}, returns the successor {@link IConnector} of a given
	 * {@code successorType}.
	 * 
	 * @param connector
	 *            {@link IConnector} for which to determine the successor.
	 * @param successorType
	 *            Type of successor {@link IConnector}.
	 * @return The successor {@link IConnector} in case it exists and is unique (i.e., the given
	 *         {@link IConnector} contains exactly one {@link IConnection} that points to a
	 *         successor {@link IConnector} of the desired type.
	 */
	@SuppressWarnings("unchecked")
	public static <T extends IConnector> T getSuccessorConnector(IConnector connector,
			Class<T> successorType) {
		T rval = null;
		for(IConnection connection : connector.getOutgoing()) {
			if(successorType.isAssignableFrom(connection.getTarget().getClass())) {
				if(rval == null) {
					rval = (T)connection.getTarget();
				} else {
					return null;
				}
			}
		}
		return null;
	}
}
