/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2011 fortiss GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.utils;

import static org.fortiss.tooling.base.layout.LayoutKeyConstants.CONNECTOR_ANGLE;
import static org.fortiss.tooling.base.utils.AngleUtils.getAngle;
import static org.fortiss.tooling.base.utils.AngleUtils.setAngle;
import static org.fortiss.tooling.base.utils.LayoutDataUtils.getNodePosition;
import static org.fortiss.tooling.base.utils.LayoutDataUtils.getNodeSize;
import static org.fortiss.tooling.base.utils.LayoutModelElementFactory.createAngle;

import org.fortiss.tooling.base.model.layout.Angle;
import org.fortiss.tooling.base.model.layout.Dimension;
import org.fortiss.tooling.base.model.layout.ILayoutedModelElement;
import org.fortiss.tooling.base.model.layout.Point;
import org.fortiss.tooling.kernel.utils.EcoreUtils;

/**
 * Utility methods for ellipse-shaped edit part layout.
 * 
 * @author hoelzl
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating GREEN Hash: 1379026092A4D8573C5CBDF0418068BD
 */
public final class EllipseLayoutUtils {
	/**
	 * Returns polar coordinates of the point on the ellipse border having the
	 * given angle.
	 * 
	 * @param angle
	 *            the angle
	 * @param ellipseSize
	 *            the wdith and height of the ellipse
	 * @param inset
	 *            the insets of the ellipse figure
	 * @return the polar coordinates
	 */
	public static Polar calculateEllipseBorderPolar(double angle, Dimension ellipseSize, int inset) {
		double a = ellipseSize.getWidth() / 2 - inset;
		double b = ellipseSize.getHeight() / 2 - inset;
		double epsilon = Math.sqrt(a * a - b * b) / a;
		double r = b / Math.sqrt(1 - epsilon * epsilon * Math.cos(angle) * Math.cos(angle));
		return new Polar(angle, r);
	}

	/**
	 * Returns Cartesian coordinates of the point on the ellipse border having
	 * the given angle.
	 * 
	 * @param angle
	 *            the angle
	 * @param ellipseSize
	 *            the wdith and height of the ellipse
	 * @param inset
	 *            the insets of the ellipse figure
	 * @return the point coordinates
	 */
	public static Point calculateEllipseBorderPoint(double angle, Dimension ellipseSize, int inset) {
		Polar polar = calculateEllipseBorderPolar(angle, ellipseSize, inset);
		return polar.toCartesian().translate(ellipseSize.getScaled(.5));
	}

	/**
	 * Layouts the given connector within the surrounding ellipse.
	 * 
	 * @param layouted
	 *            the layouted connector model element
	 * @param container
	 *            the container ellipse
	 * @param location
	 *            the target point to be moved to the ellipse border
	 */
	public static void layoutConnectorInEllipse(ILayoutedModelElement layouted,
			ILayoutedModelElement container, Point location) {
		Point position = EcoreUtils.copy(location);
		Point np = getNodePosition(container);
		position.translate(-np.getX(), -np.getY());

		Dimension nd = getNodeSize(container);
		Dimension d =
				LayoutModelElementFactory.createDimension(nd.getWidth(), nd.getHeight(),
						nd.toString() + "_Dimension");
		position.translate(d.getScaled(-.5));

		// store the value in the model
		setConnectorAngle(layouted, new Polar(position).angle);
	}

	/** Sets the layouted connector's angle. */
	public static void setConnectorAngle(ILayoutedModelElement layouted, double angle) {
		setAngle(layouted, CONNECTOR_ANGLE, angle);
	}

	/** Returns the layouted connector's angle. */
	public static Angle getConnectorAngle(ILayoutedModelElement layouted) {
		Angle angle = getAngle(layouted, CONNECTOR_ANGLE);
		if(angle == null) {
			return createAngle(0.0, CONNECTOR_ANGLE);
		}
		return angle;
	}

	/**
	 * Class for storing polar coordinates.
	 * 
	 * The angles are interpreted clockwise in the following way: 0deg: East;
	 * 90deg(pi/2): South; 180deg(pi): West; 270deg(3/2pi): North
	 * 
	 * @author wandinger
	 * @author hummel
	 * @author $Author$
	 * @version $Rev$
	 * @levd.rating YELLOW Rev: 519
	 */
	public static class Polar {

		/** The angle */
		public double angle;

		/** The distance */
		public double distance;

		/** The constructor */
		public Polar(double angle, double distance) {
			this.angle = angle;
			this.distance = distance;
		}

		/** Create new polar values from cartesian ones. */
		public Polar(Point pos) {
			angle = Math.atan2(pos.getY(), pos.getX());
			distance = Math.sqrt(pos.getX() * pos.getX() + pos.getY() * pos.getY());
		}

		/** Returns a cartesian representation. */
		public Point toCartesian() {
			int x = (int)(distance * Math.cos(angle));
			int y = (int)(distance * Math.sin(angle));
			return LayoutModelElementFactory.createPoint(x, y, this.toString() + "_Polar");
		}
	}
}
