/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2016 fortiss GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.kernel.ui.internal.introspection.details.handler;

import org.eclipse.jface.action.Action;
import org.eclipse.jface.action.IMenuListener;
import org.eclipse.jface.action.IMenuManager;
import org.eclipse.jface.action.MenuManager;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.jface.viewers.ITableLabelProvider;
import org.eclipse.jface.viewers.ITreeContentProvider;
import org.eclipse.jface.viewers.TreeViewer;
import org.eclipse.jface.viewers.ViewerSorter;
import org.eclipse.swt.SWT;
import org.eclipse.swt.custom.CTabFolder;
import org.eclipse.swt.custom.CTabItem;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Tree;
import org.eclipse.ui.dialogs.FilteredTree;
import org.eclipse.ui.dialogs.PatternFilter;
import org.fortiss.tooling.kernel.ui.internal.introspection.details.DetailsUIHandlerBase;

/**
 * Base class for details UI implementations with filtered tree viewer.
 * 
 * @author hoelzl
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating GREEN Hash: E254CABB1D9CD30A37F1ED1823E47A62
 */
public abstract class IntrospectionDetailsUIHandlerBase extends DetailsUIHandlerBase {

	/** Creates only the filtered tree control. */
	protected final Composite createFilteredTree(Composite parent) {
		return createFilteredTree(parent, null, null);
	}

	/** Creates the filtered tree control. */
	protected final Composite createFilteredTree(Composite parent, String headingsLabel,
			String footerLabel) {
		Composite composite = new Composite(parent, SWT.BORDER);
		composite.setLayout(new GridLayout(1, true));

		if(headingsLabel != null) {
			Label heading = new Label(composite, SWT.NONE);
			heading.setText(headingsLabel);
			heading.setLayoutData(new GridData(SWT.BEGINNING, SWT.CENTER, true, false));
		}

		FilteredTree filteredTree =
				new FilteredTree(composite, SWT.BORDER, createPatternFilter(), true);

		TreeViewer treeViewer = filteredTree.getViewer();
		Tree tree = treeViewer.getTree();
		tree.setHeaderVisible(true);
		tree.setLinesVisible(true);

		createTreeColumns(tree);

		treeViewer.setContentProvider(createContentProvider());
		treeViewer.setLabelProvider(createLabelProvider());
		treeViewer.setSorter(createSorter());
		treeViewer.setInput(getInputObject());

		addContextMenu(treeViewer);

		if(footerLabel != null) {
			Label lbl = new Label(composite, SWT.BOLD);
			lbl.setText(footerLabel);
			lbl.setLayoutData(new GridData(SWT.BEGINNING, SWT.CENTER, true, false));
		}

		return composite;
	}

	/**
	 * Creates the context menu by delegating menu entry creation
	 * to {@link #populateContextMenu(Object, MenuManager)} and {@link #testSelection(Object)}.
	 */
	private void addContextMenu(final TreeViewer treeViewer) {
		final MenuManager mgr = new MenuManager();
		mgr.setRemoveAllWhenShown(true);
		mgr.addMenuListener(new IMenuListener() {

			@Override
			public void menuAboutToShow(IMenuManager manager) {
				IStructuredSelection sel = (IStructuredSelection)treeViewer.getSelection();
				if(sel != null && !sel.isEmpty() && testSelection(sel.getFirstElement())) {
					populateContextMenu(sel.getFirstElement(), mgr);
				}
			}
		});
		treeViewer.getControl().setMenu(mgr.createContextMenu(treeViewer.getControl()));
	}

	/**
	 * Populates the context menu. May assume {@link #testSelection(Object)} returned {@code true}.
	 * The default does not create any menu entry.
	 */
	protected void populateContextMenu(@SuppressWarnings("unused") Object selection,
			@SuppressWarnings("unused") MenuManager mgr) {
		// create nothing
	}

	/** Creates an action with the given name and selection handler. */
	protected final Action createAction(String name, Runnable runner) {
		return new Action(name) {
			/** {@inheritDoc} */
			@Override
			public void run() {
				runner.run();
			}
		};
	}

	/** Creates the copy class name action for the given class. */
	protected final Action createCopyClassNameAction(Class<?> clazz) {
		return createAction("Copy Qualified Class Name: " + clazz.getSimpleName(),
				new CopyClassNameToClipBoardRunnable(clazz.getName()));
	}

	/**
	 * Tests the given selected object for validity to show context menu. The default returns
	 * {@code false}.
	 */
	protected boolean testSelection(@SuppressWarnings("unused") Object selection) {
		return false;
	}

	/** Creates a tab folder with the filtered tree as a tab. */
	protected final CTabFolder createTabFolder(Composite parent, String... tabNames) {
		CTabFolder tabFolder = new CTabFolder(parent, SWT.BOTTOM);
		if(tabNames != null && tabNames.length > 0) {
			for(String tName : tabNames) {
				CTabItem filterTreeTab = new CTabItem(tabFolder, SWT.NULL);
				filterTreeTab.setText(tName);
			}
			tabFolder.setSelection(tabFolder.getItem(0));
		}
		return tabFolder;
	}

	/** Creates the filtered tree inside the tab folder. */
	protected final CTabFolder createFilteredTreeInTabFolder(Composite parent,
			String headingsLabel, String footerLabel, String... tabNames) {
		CTabFolder tabFolder = createTabFolder(parent, tabNames);
		Composite c = createFilteredTree(tabFolder, headingsLabel, footerLabel);
		tabFolder.getItem(0).setControl(c);
		return tabFolder;
	}

	/** Create the columns of the tree. */
	protected abstract void createTreeColumns(Tree tree);

	/** Creates the content provider. */
	protected abstract ITreeContentProvider createContentProvider();

	/** Creates the label provider. */
	protected abstract ITableLabelProvider createLabelProvider();

	/** Creates the pattern filter. */
	protected abstract PatternFilter createPatternFilter();

	/** Returns the input object for the tree viewer. */
	protected abstract Object getInputObject();

	/**
	 * Creates the sorter for the filtered tree. The default is {@link ViewerSorter#ViewerSorter()}.
	 */
	protected ViewerSorter createSorter() {
		return new ViewerSorter();
	}
}
