/*-----------------------------------------------------------------------+
 | org.fortiss.tooling.kernel.ui.internal.views
 |                                                                       |
   $Id$            
 |                                                                       |
 | Copyright (c)  2016-today fortiss GmbH                                |
 |                                                                       |
 +-----------------------------------------------------------------------*/
package org.fortiss.tooling.kernel.ui.internal.views;

import static org.eclipse.jface.dialogs.MessageDialog.openError;
import static org.eclipse.jface.resource.ImageDescriptor.createFromImage;

import java.util.Collections;
import java.util.List;
import java.util.stream.Collectors;

import org.eclipse.emf.ecore.EObject;
import org.eclipse.jface.action.Action;
import org.eclipse.jface.action.ActionContributionItem;
import org.eclipse.jface.action.IAction;
import org.eclipse.jface.action.IContributionItem;
import org.eclipse.jface.action.MenuManager;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.jface.resource.ImageDescriptor;
import org.eclipse.jface.viewers.DecorationOverlayIcon;
import org.eclipse.jface.viewers.IDecoration;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Shell;
import org.fortiss.tooling.kernel.model.constraints.ConstrainedWithChecksum;
import org.fortiss.tooling.kernel.model.constraints.ConstraintInstance;
import org.fortiss.tooling.kernel.model.constraints.ErrorVerificationStatus;
import org.fortiss.tooling.kernel.model.constraints.FailVerificationStatus;
import org.fortiss.tooling.kernel.model.constraints.IConstrained;
import org.fortiss.tooling.kernel.model.constraints.IConstraintVerificationStatus;
import org.fortiss.tooling.kernel.model.constraints.OutdatedVerificationStatus;
import org.fortiss.tooling.kernel.model.constraints.SuccessVerificationStatus;
import org.fortiss.tooling.kernel.service.ICommandStackService;
import org.fortiss.tooling.kernel.service.IConstraintVerificationService;
import org.fortiss.tooling.kernel.service.IConstraintVerificationService.IFix;
import org.fortiss.tooling.kernel.ui.ESharedImages;
import org.fortiss.tooling.kernel.ui.ToolingKernelUIActivator;
import org.fortiss.tooling.kernel.ui.extension.IContextMenuContributor;
import org.fortiss.tooling.kernel.ui.extension.IModelElementHandler;
import org.fortiss.tooling.kernel.ui.extension.data.ContextMenuContextProvider;
import org.fortiss.tooling.kernel.ui.service.IConstraintVerificationUIService;
import org.fortiss.tooling.kernel.ui.service.IContextMenuService;
import org.fortiss.tooling.kernel.ui.service.IModelEditorBindingService;
import org.fortiss.tooling.kernel.ui.service.IModelElementHandlerService;

/**
 * Context menu for constraints.
 * 
 * @author hoelzl
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating YELLOW Hash: B46F0F098D084B44B1BE8D9C772D9058
 */
public class ConstraintMenu implements IContextMenuContributor {

	/** {@inheritDoc} */
	@Override
	public List<IContributionItem> getContributedItems(EObject selectedObject,
			ContextMenuContextProvider contextProvider) {
		if(!(selectedObject instanceof IConstrained)) {
			return Collections.emptyList();
		}
		return ((IConstrained)selectedObject).getConstraints().stream()
				.filter(c -> shouldGetMenuEntry(c))
				.map(c -> constraintToAction(c, (IConstrained)selectedObject))
				.collect(Collectors.toList());
	}

	/**
	 * @param c
	 * @return <code>true</code> iff <code>c</code> should get a menu entry, i.e., if not successful
	 *         and activated.
	 */
	private boolean shouldGetMenuEntry(ConstraintInstance c) {
		IConstraintVerificationStatus status =
				IConstraintVerificationUIService.getInstance().getStatus(c);
		return !(status == null || status instanceof SuccessVerificationStatus);
	}

	/**
	 * @param c
	 * @param elt
	 * @return The submenu or menu entry corresponding to the status of <code>c</code>.
	 *         <code>elt</code> is needed to prevent displaying an entry to go to <code>elt</code>
	 *         (not hurtful, but useless, since the user is already seeing it).
	 */
	private IContributionItem constraintToAction(ConstraintInstance c, IConstrained elt) {
		IConstraintVerificationStatus status =
				IConstraintVerificationUIService.getInstance().getStatus(c);
		ConstraintSubMenuBase m;
		if(status instanceof FailVerificationStatus) {
			m = new CheckFailingConstraintAction(c, elt);
		} else if(status instanceof ErrorVerificationStatus) {
			m = new CheckErrorConstraintAction(c, elt);
		} else if(status instanceof OutdatedVerificationStatus) {
			m = new CheckOutdatedConstraintAction(c, elt);
		} else {
			// should not happen
			return null;
		}
		if(m.getItems().length > 1) {
			return m;
		}
		IAction uniqueAction = m.getTopActionContribution().getAction();
		String txt = m.getMenuText();
		String newText = txt + " -> " + uniqueAction.getText();
		uniqueAction.setText(newText);
		uniqueAction.setImageDescriptor(m.getImageDescriptor());
		return m.getTopActionContribution();
	}

	/** {@inheritDoc} */
	@Override
	public String getMenuSectionID() {
		return IContextMenuService.BEFORE_GLOBAL_MENU_SECTION_ID;
	}

	/** Action to "go to" a given constrained element. */
	private static class GoToConstrained extends Action {

		/** The constrained element to open. */
		private IConstrained c;

		/** Constructor. */
		public GoToConstrained(IConstrained c) {
			super("Go to " + IModelElementHandlerService.getInstance().getName(c),
					createFromImage(IModelElementHandlerService.getInstance().getIcon(c)));
			this.c = c;
		}

		/** {@inheritDoc} */
		@Override
		public void run() {
			IModelElementHandler<EObject> handler =
					IModelElementHandlerService.getInstance().getModelElementHandler(c);
			EObject eltToOpen = handler == null ? c : handler.handleOpenModelElementRequest(c);
			IModelEditorBindingService.getInstance().openInEditor(eltToOpen);
		}
	}

	/** Get the icon of the prototype. */
	public static String getText(ConstraintInstance c, String suffix) {
		String id = c.getConstraintID();
		String desc = IConstraintVerificationUIService.getInstance().getDescription(id);
		return "\"" + desc + "\": " + suffix;
	}

	/** Get the icon of the prototype. */
	public static ImageDescriptor getIcon(ConstraintInstance c, ImageDescriptor overlay) {
		String id = c.getConstraintID();
		ImageDescriptor imgd =
				IConstraintVerificationUIService.getInstance().getIconImageDescriptor(id);
		Image img = imgd == null ? null : imgd.createImage();
		if(img == null) {
			return null;
		}
		ImageDescriptor[] descriptors = new ImageDescriptor[5];
		descriptors[IDecoration.BOTTOM_LEFT] = overlay;
		return new DecorationOverlayIcon(img, descriptors);
	}

	/** Action for creating a new prototype. */
	private static class ConstraintSubMenuBase extends MenuManager {

		/** The constraint. */
		protected ConstraintInstance c;

		/**
		 * The action to get more information about the constraint status.
		 * The same action allows to update an outdated constraint status (there is no
		 * "more information" to get out of an outdated constraint); it is stored in a field to
		 * allow the "update" action to change the icon and the text going with it.
		 */
		protected ActionContributionItem moreInfoAction;

		/** Constructor. */
		public ConstraintSubMenuBase(ConstraintInstance c, IConstrained selectedElt, String suffix,
				ImageDescriptor overlay) {
			super(getText(c, suffix), getIcon(c, overlay), null);
			this.c = c;
			moreInfoAction = new ActionContributionItem(new OpenStatusAction());
			this.add(moreInfoAction);
			for(ConstrainedWithChecksum cwc : c.getConstrainedsWithChecksum()) {
				if(!selectedElt.equals(cwc.getConstrained())) {
					this.add(new ActionContributionItem(new GoToConstrained(cwc.getConstrained())));
				}
			}
		}

		/**
		 * @return The top action contribution. Useful in case we get aware that there is only one
		 *         contribution and we therefore want to present it without going through a submenu.
		 */
		public ActionContributionItem getTopActionContribution() {
			return moreInfoAction;
		}

		/** Action to update a constraint. */
		protected class OpenStatusAction extends Action {

			/** Constructor. */
			public OpenStatusAction() {
				super("More information...", ToolingKernelUIActivator
						.getImageDescriptor("icons/info.gif"));
			}

			/** {@inheritDoc} */
			@Override
			public void run() {
				IConstraintVerificationStatus status =
						IConstraintVerificationUIService.getInstance().getStatus(c);
				if(status instanceof OutdatedVerificationStatus &&
						IConstraintVerificationUIService.getInstance().canOpen(c)) {
					ICommandStackService.getInstance().runAsCommand(c, new Runnable() {
						@Override
						public void run() {
							IConstraintVerificationUIService.getInstance().openStatus(c);
						}
					});
					return;
				}
				if(IConstraintVerificationUIService.getInstance().canOpen(c)) {
					IConstraintVerificationUIService.getInstance().openStatus(c);
				} else {
					// In the very rare cases where the status cannot be open (if the defaults are
					// used, this should not happen), we provide some defaults.
					Shell sh = Display.getCurrent().getActiveShell();
					String name = ConstraintMenu.getText(c, "");
					if(status instanceof FailVerificationStatus) {
						String msg = (name == null ? "The " : name);
						msg += " is not satisfied.";
						MessageDialog.openInformation(sh, "Unsatisfied constraint", msg);
					} else if(status instanceof ErrorVerificationStatus) {
						String msg = "There was an error while verifying ";
						msg += (name == null ? "the " : "") + name + ".";
						openError(sh, "Error while verifying constraint", msg);
					} else if(status instanceof OutdatedVerificationStatus) {
						String msg = (name == null ? "The " : name) + " is outdated.";
						MessageDialog.openWarning(sh, "Outdated constraint", msg);
					}
				}
			}
		}
	}

	/** Action for creating a new prototype. */
	private static class CheckOutdatedConstraintAction extends ConstraintSubMenuBase {

		/** Constructor. */
		public CheckOutdatedConstraintAction(ConstraintInstance c, IConstrained selectedElt) {
			super(c, selectedElt, "Outdated", ESharedImages.WARNING_OVERLAY.getImageDescriptor());
			IAction action = moreInfoAction.getAction();
			action.setImageDescriptor(ToolingKernelUIActivator.getImageDescriptor("icons/ok.png"));
			action.setText("Check");
		}
	}

	/** Action for creating a new prototype. */
	private static class CheckUnsuccessfulConstraintAction extends ConstraintSubMenuBase {

		/** Constructor. */
		public CheckUnsuccessfulConstraintAction(ConstraintInstance c, IConstrained selectedElt,
				String suffix) {
			super(c, selectedElt, suffix, getOverlay(c));
			List<IFix> fixes = IConstraintVerificationService.getInstance().fixes(c);
			if(fixes != null) {
				for(IFix fix : fixes) {
					this.add(new ActionContributionItem(new FixAction(fix)));
				}
			}
		}

		/** Get the correct overlay for the given constraint. */
		private static ImageDescriptor getOverlay(ConstraintInstance c) {
			if(IConstraintVerificationUIService.getInstance().shallDisplayAsWarning(c)) {
				return ESharedImages.WARNING_OVERLAY.getImageDescriptor();
			}
			return ESharedImages.ERROR_OVERLAY.getImageDescriptor();
		}

		/** Action to update a constraint. */
		private class FixAction extends Action {

			/** The fix to achieve. */
			IFix fix;

			/** Constructor. */
			public FixAction(IFix fix) {
				super(fix.getDescription(), ToolingKernelUIActivator
						.getImageDescriptor("icons/fix.png"));
				this.fix = fix;
			}

			/** {@inheritDoc} */
			@Override
			public void run() {
				ICommandStackService.getInstance().runAsCommand(c, new Runnable() {

					@Override
					public void run() {
						fix.runFix(IConstraintVerificationUIService.getInstance().getStatus(c));
					}
				});
			}
		}
	}

	/** Action for creating a new prototype. */
	private static class CheckFailingConstraintAction extends CheckUnsuccessfulConstraintAction {

		/** Constructor. */
		public CheckFailingConstraintAction(ConstraintInstance c, IConstrained selectedElt) {
			super(c, selectedElt, "Unsatisfied");
		}
	}

	/** Action for creating a new prototype. */
	private static class CheckErrorConstraintAction extends CheckUnsuccessfulConstraintAction {

		/** Constructor. */
		public CheckErrorConstraintAction(ConstraintInstance c, IConstrained selectedElt) {
			super(c, selectedElt, "Error encountered");
		}
	}
}
