/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2016 fortiss GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.kernel.extension.base;

import java.util.ArrayList;
import java.util.List;

import org.eclipse.emf.ecore.EObject;
import org.fortiss.tooling.kernel.extension.IConstraintVerifier;
import org.fortiss.tooling.kernel.model.constraints.ConstraintInstance;
import org.fortiss.tooling.kernel.model.constraints.IConstrained;
import org.fortiss.tooling.kernel.model.constraints.IConstraintVerificationStatus;
import org.fortiss.tooling.kernel.service.IConstraintVerificationService.IFix;
import org.fortiss.tooling.kernel.utils.ConstraintsUtils;

/**
 * Base for {@link IConstraintVerifier}.
 * 
 * @author aravantinos
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating GREEN Hash: 470ADF281E0F28CCCE27720EAB2A0E45
 */
public class ConstraintVerifierBases {
	/** Most generic base. */
	public abstract static class ConstraintVerifierBase implements IConstraintVerifier {
		/** {@inheritDoc} */
		@Override
		public void removeConstraintIrrelevantContent(EObject obj) {
			// By default, we do nothing. Should be adapted in general but makes the first
			// developments of a verifier easier.
		}

		/** {@inheritDoc} */
		@Override
		public void cancel(ConstraintInstance constraint) {
			// By default, we do nothing. Not all constraints are so heavy that they deserve to have
			// a cancellation procedure.
		}

		/** {@inheritDoc} */
		@Override
		public List<IFix>
				fixes(ConstraintInstance constraint, IConstraintVerificationStatus status) {
			// No fix by default
			return new ArrayList<IFix>();
		}

		/** {@inheritDoc} */
		@Override
		public ConstraintInstance addConstraintIfNeeded(IConstrained constrained) {
			if(ConstraintsUtils.getConstraintOfType(constrained, getID()) != null) {
				return null;
			}
			ConstraintInstance c = createConstraintIfNeeded(constrained);
			if(c != null) {
				ConstraintsUtils.createOutdatedVerificationStatus(c);
			}
			return c;
		}

		/**
		 * Creates and return a constraint on the provided element.
		 * 
		 * Subclasses shall return <code>null</code> if the provided element should not be
		 * constrained (whatever is the reason: the constraint verifier is only judge here).
		 */
		public abstract ConstraintInstance createConstraintIfNeeded(IConstrained constrained);
	}

	/**
	 * Base for a constraint verifier which is as close as possible to the old constraint system:
	 * - the constrained element contains all the information necessary for the verification
	 * (nothing stored in the constraint)
	 * - only one element is constrained
	 */
	public static abstract class ConstraintCheckerBase extends ConstraintVerifierBase {
		/** {@inheritDoc} */
		@Override
		public final IConstraintVerificationStatus verify(ConstraintInstance constraint) {
			return verify(ConstraintsUtils.getConstrained(constraint));
		}

		/** Verify the given constrained element. */
		public abstract IConstraintVerificationStatus verify(IConstrained constrained);

		/**
		 * Determines whether the constraint verified by this verifier is applicable to the given
		 * model element.
		 */
		public abstract boolean isApplicable(IConstrained constrained);

		/** {@inheritDoc} */
		@Override
		public ConstraintInstance addConstraintIfNeeded(IConstrained constrained) {
			return isApplicable(constrained) ? super.addConstraintIfNeeded(constrained) : null;
		}
	}
}
