/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2011 fortiss GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.kernel.utils;

import java.util.NoSuchElementException;
import java.util.function.Consumer;

import org.fortiss.tooling.kernel.model.constraints.ConstraintInstance;
import org.fortiss.tooling.kernel.model.constraints.ConstraintsFactory;
import org.fortiss.tooling.kernel.model.constraints.ErrorVerificationStatus;
import org.fortiss.tooling.kernel.model.constraints.FailVerificationStatus;
import org.fortiss.tooling.kernel.model.constraints.IConstrained;
import org.fortiss.tooling.kernel.model.constraints.IConstraintVerificationStatus;
import org.fortiss.tooling.kernel.model.constraints.OutdatedVerificationStatus;
import org.fortiss.tooling.kernel.model.constraints.SuccessVerificationStatus;
import org.fortiss.tooling.kernel.service.IConstraintVerificationService;
import org.fortiss.tooling.kernel.service.IConstraintVerificationService.IFix;

/**
 * Utility methods for dealing with constraints.
 * 
 * @author aravantinos
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating YELLOW Hash: 24089C9BFEF4DA99D78CDE0C2DCE2996
 */
public class ConstraintsUtils {
	/**
	 * @param c
	 * @param id
	 * @return <code>true</code> if <code>c</code> contains a constraint of constraint type ID
	 *         <code>id</code>.
	 */
	public static boolean containsConstraintOfType(IConstrained c, String id) {
		return getConstraintOfType(c, id) != null;
	}

	/**
	 * @param c
	 * @param id
	 * @return the first constraint of <code>c</code> of constraint type ID <code>id</code>.
	 */
	public static ConstraintInstance getConstraintOfType(IConstrained c, String id) {
		try {
			return c.getConstraints().stream().filter(x -> x.getConstraintID().equals(id))
					.findFirst().get();
		} catch(NoSuchElementException e) {
			return null;
		}
	}

	/** Exception which embeds an error. */
	public static class ErrorEmbeddingException extends RuntimeException {

		/** The embedded error. */
		public ErrorVerificationStatus error;

		/** Constructor. */
		public ErrorEmbeddingException(ErrorVerificationStatus error) {
			super();
			this.error = error;
		}
	}

	/**
	 * @param c
	 * @return A "success" verification status for <code>c</code>.
	 */
	public static SuccessVerificationStatus createSuccessVerificationStatus(ConstraintInstance c) {
		SuccessVerificationStatus s =
				ConstraintsFactory.eINSTANCE.createSuccessVerificationStatus();
		c.setVerificationStatus(s);
		return s;
	}

	/**
	 * @param c
	 * @return A "outdated" verification status for <code>c</code>.
	 */
	public static OutdatedVerificationStatus createOutdatedVerificationStatus(ConstraintInstance c) {
		OutdatedVerificationStatus status;
		if(c.getVerificationStatus() instanceof OutdatedVerificationStatus) {
			status = (OutdatedVerificationStatus)c.getVerificationStatus();
		} else {
			status = ConstraintsFactory.eINSTANCE.createOutdatedVerificationStatus();
			c.setVerificationStatus(status);
		}
		return status;
	}

	/**
	 * @param c
	 * @return A "fail" verification status for <code>c</code>.
	 */
	public static FailVerificationStatus createFailVerificationStatus(ConstraintInstance c) {
		FailVerificationStatus status = ConstraintsFactory.eINSTANCE.createFailVerificationStatus();
		c.setVerificationStatus(status);
		return status;
	}

	/**
	 * @param c
	 * @return An "error" verification status for <code>c</code>.
	 */
	public static ErrorVerificationStatus createErrorVerificationStatus(ConstraintInstance c) {
		ErrorVerificationStatus status =
				ConstraintsFactory.eINSTANCE.createErrorVerificationStatus();
		c.setVerificationStatus(status);
		return status;
	}

	/**
	 * @param c
	 * @return the first element constrained by <code>c</code>
	 */
	public static IConstrained getFirstConstrained(ConstraintInstance c) {
		// get(0) because the constraint shall contain at least one constrained element
		// (responsibility of the caller!) and because we want the first element
		return c.getConstrainedsWithChecksum().get(0).getConstrained();
	}

	/**
	 * @param constraint
	 * @param constrained
	 *            Sets the constrained element of <code>constraint</code>
	 */
	public static void setFirstConstrained(ConstraintInstance constraint, IConstrained constrained) {
		IConstraintVerificationService.getInstance().setConstrainedElement(constraint, constrained,
				0);
	}

	/**
	 * @param c
	 * @return the second element constrained by <code>c</code>
	 */
	public static IConstrained getSecondConstrained(ConstraintInstance c) {
		// get(1) because the constraint shall contain at least one constrained element
		// (responsibility of the caller!) and because we want the first element
		return c.getConstrainedsWithChecksum().get(1).getConstrained();
	}

	/**
	 * @param constraint
	 * @param constrained
	 *            Sets the constrained element of <code>constraint</code>
	 */
	public static void
			setSecondConstrained(ConstraintInstance constraint, IConstrained constrained) {
		IConstraintVerificationService.getInstance().setConstrainedElement(constraint, constrained,
				1);
	}

	/**
	 * @param c
	 * @return the element constrained by <code>constraint</code>.
	 *         Use this function if your constraint constrains only one element!
	 */
	public static IConstrained getConstrained(ConstraintInstance c) {
		return getFirstConstrained(c);
	}

	/**
	 * @param constraint
	 * @param constrained
	 *            Sets the constrained element of <code>constraint</code>.
	 *            Use this function if your constraint constrains only one element!
	 */
	public static void setConstrained(ConstraintInstance constraint, IConstrained constrained) {
		IConstraintVerificationService.getInstance().setConstrainedElement(constraint, constrained,
				0);
	}

	/**
	 * @param id
	 * @param constrained
	 * @return a constraint of ID <code>id</code> constraining <code>constrained</code>
	 */
	public static ConstraintInstance createConstraint(String id, IConstrained constrained) {
		ConstraintInstance c = ConstraintsFactory.eINSTANCE.createConstraintInstance();
		c.setConstraintID(id);
		ConstraintsUtils.setConstrained(c, constrained);
		createOutdatedVerificationStatus(c);
		return c;
	}

	/**
	 * @param id
	 * @param constrained1
	 * @param constrained2
	 * @return a constraint of ID <code>id</code> constraining <code>constrained1</code> and
	 *         <code>constrained2</code>.
	 */
	public static ConstraintInstance createConstraint(String id, IConstrained constrained1,
			IConstrained constrained2) {
		ConstraintInstance c = ConstraintsFactory.eINSTANCE.createConstraintInstance();
		c.setConstraintID(id);
		ConstraintsUtils.setFirstConstrained(c, constrained1);
		ConstraintsUtils.setSecondConstrained(c, constrained2);
		createOutdatedVerificationStatus(c);
		return c;
	}

	/** Base for fixes. */
	public static class FixBase implements IFix {
		/** The description. */
		private String description;

		/** The fix itself. */
		private Consumer<IConstraintVerificationStatus> fix;

		/** Constructor. */
		public FixBase(String description, Consumer<IConstraintVerificationStatus> fix) {
			this.description = description;
			this.fix = fix;
		}

		/** {@inheritDoc} */
		@Override
		public String getDescription() {
			return description;
		}

		/** {@inheritDoc} */
		@Override
		public void runFix(IConstraintVerificationStatus status) {
			fix.accept(status);
		}
	}
}
